#include <stdint.h>
#include <stdio.h>

#include "libsat.h"

#define	SAT_SIGN_MASK	0x80000000
#define	SAT_HIGH_PART	0xffff0000
#define	SAT_LOW_PART	0x0000ffff

#define Q_N	16
#define Q_SCALE	(1 << Q_N)

#define	overflow_if_true(exp) {\
	if (exp) {\
		sat_flags |= SAT_OVERFLOW;\
		res = SAT_MAX_VALUE;\
		goto overflow;\
	}\
}

static int
lb(int n)
{
	int x, c;

	x = 0;
	c = 1;

	while (c < n) {
		c <<= 1;
		++x;
	}

	return x;
}

int32_t
sat_add(int32_t a, int32_t b)
{
	int32_t res;

	res = a + b;
	if (((a ^ res) & (b ^ res)) < 0) {
		if (res < 0) {
			res =  SAT_MAX_VALUE;
		} else {
			res =  SAT_MIN_VALUE;
		}

		sat_flags |= SAT_OVERFLOW;
	}

	return res;
}

int32_t
sat_sub(int32_t a, int32_t b)
{
	return sat_add(a, -b);
}

int32_t
sat_mul(int32_t a, int32_t b)
{
	uint32_t sign, ax, bx, tmp;
	int32_t res;

	sign = (a ^ b) & SAT_SIGN_MASK;

	if (a < 0)
		a = -a;

	if (b < 0)
		b = -b;

	ax = a & SAT_LOW_PART;
	bx = b & SAT_LOW_PART;
	a >>= 16;
	b >>= 16;

	tmp = a * b;
	overflow_if_true(tmp >= SAT_LOW_PART)

	res = tmp << 16;
	tmp = a * bx;
	overflow_if_true(tmp & SAT_SIGN_MASK)

	res += tmp;
	overflow_if_true(res & SAT_SIGN_MASK)

	tmp = ax * b;
	overflow_if_true(tmp & SAT_SIGN_MASK)

	res += tmp;
	overflow_if_true(res & SAT_SIGN_MASK)

	tmp = ax * bx;
	res += ((tmp >> 15) & 0x1);
	res += (tmp >> 16);
	overflow_if_true(res & SAT_SIGN_MASK)

overflow:
	if (sign != 0)
		res = -res;

	return res;
}

int32_t
sat_div(int32_t a, int32_t b)
{
	uint32_t  remainder, mask, sign, tmp;
	int32_t res;

	sign = (a ^ b) & SAT_SIGN_MASK;

	if (a < 0)
		a = -a;

	if (b < 0)
		b = -b;

	if (b == 0) {
		if (a == 0)
			res = SAT_ONE;
		else
			res = SAT_MAX_VALUE;
		sat_flags |= SAT_BY_ZERO;
		goto overflow;
	}

	/* get high part of result without rounding */
	tmp = a / b;

	/* round up to half */
	remainder = (a % b) + (b >> 17);
	if (remainder >= b) {
		remainder -= b;
		tmp++;
	}

	overflow_if_true(tmp >= SAT_LOW_PART)
	res = tmp;

	/* get low part from remainder and rounding tail */
	a = b << 15;
	mask = SAT_SIGN_MASK;
	while (mask > SAT_LOW_PART) {
		remainder <<= 1;
		res <<= 1;

		if ((a & mask) != 0)
			remainder |= 0x1;

		if (remainder >= b) {
			res |= 0x1;
			remainder -= b;
		}

		mask >>= 1;
	}


overflow:
	if (sign != 0)
		res = -res;

	return res;
}

int32_t
sat_sin(int32_t x)
{
	int32_t tmp, res, sign, tx;

	sign = x & SAT_SIGN_MASK;
	if (sign < 0)
		x = -x;

	tmp = sat_div(x, SAT_PI);

	tmp &= SAT_HIGH_PART;
	if (tmp & 0x00010000)
		sign ^= SAT_SIGN_MASK;

	tmp = sat_mul(tmp, SAT_PI);
	x = sat_sub(x, tmp);
	if (x > SAT_PI2)
		x = sat_sub(SAT_PI, x);

	res = x;
	tmp = x;
	x = sat_mul(x, x);

	tmp = sat_mul(tmp, x);
	tx  = sat_mul(tmp, 0xffffd556);
	res = sat_add(res, tx);

	tmp = sat_mul(tmp, x);
	tx  = sat_mul(tmp, 0x00000223);
	res = sat_add(res, tx);

	tmp = sat_mul(tmp, x);
	tx  = sat_mul(tmp, 0xfffffff3);
	res = sat_add(res, tx);


	if (sign < 0)
		res = -res;

	return res;
}

int32_t
sat_cos(int32_t x)
{
	int32_t tmp, res, sign, tx;

	sign = 0;
	if (x < 0)
		x = -x;

	tmp = sat_div(x, SAT_PI);

	tmp &= SAT_HIGH_PART;
	if (tmp & 0x00010000)
		sign ^= SAT_SIGN_MASK;

	tmp = sat_mul(tmp, SAT_PI);
	x = sat_sub(x, tmp);
	if (x > SAT_PI2) {
		x = sat_sub(SAT_PI, x);
		sign ^= SAT_SIGN_MASK;
	}

	res = SAT_ONE;
	x = sat_mul(x, x);

	tmp = x;
	tx  = sat_mul(tmp, 0xffff7fff);
	res = sat_add(res, tx);

	tmp = sat_mul(tmp, x);
	tx  = sat_mul(tmp, 0x00000aa9);
	res = sat_add(res, tx);

	tmp = sat_mul(tmp, x);
	tx  = sat_mul(tmp, 0xffffffa5);
	res = sat_add(res, tx);

	tmp = sat_mul(tmp, x);
	tx  = sat_mul(tmp, 0x00000002);
	res = sat_add(res, tx);

	if (sign < 0)
		res = -res;

	return res;
}

int
sat_fft(int32_t *x, int n, int inv)
{
	int32_t tre, tim, wr, wi, sr, si, arg, tmp;
	int32_t *f0, *f2, *f0r, *f0i, *f2r, *f2i;
	int stages, frame, framed2;
	int i, j, k;

	if ((n > 16384) || (n < 0))
		return SAT_TOO_LONG;

	for (i = 0; i < n; i++) {
		unsigned int cnt, a;
		int32_t *p, *p2, tmp;

		a = 0;
		for (cnt = 1; cnt < n; cnt <<= 1) {
			a <<= 1;
			if (i & cnt)
				a |= 0x1;
		}

		if (a > i) {
			p = x + (a << 1);
			p2 = x + (i << 1);
			tmp = *p;
			*p++ = *p2;
			*p2++ = tmp;
			tmp = *p;
			*p = *p2;
			*p2 = tmp;
		}
	}

	stages = lb(n);
	frame = 1;

	for (i = 0; i < stages; i++) {
		framed2 = frame;
		frame <<= 1;

		f0 = x;
		f2 = x + frame;

		arg = framed2 << 16;
		arg = sat_div(SAT_PI, arg);
		wr = sat_cos(arg);
		wi = sat_sin(arg);
		if (inv == 0)
			wi = -wi;

		for (j = 0; j < n; j += frame) {

			sr = SAT_ONE;
			si = SAT_ZERO;

			for (k = 0; k < framed2; k++) {
				f0r = f0++;
				f0i = f0++;
				f2r = f2++;
				f2i = f2++;

				tre = sat_mul(*f2r, sr);
				tmp = sat_mul(*f2i, si);
				tre = sat_sub(tre, tmp);

				tim = sat_mul(*f2i, sr);
				tmp = sat_mul(*f2r, si);
				tim = sat_add(tim, tmp);

				*f2r = sat_sub(*f0r, tre);
				*f2i = sat_sub(*f0i, tim);
				*f0r = sat_add(*f0r, tre);
				*f0i = sat_add(*f0i, tim);

				tre = sat_mul(sr, wr);
				tmp = sat_mul(si, wi);
				tre = sat_sub(tre, tmp);

				si  = sat_mul(si, wr);
				tmp = sat_mul(sr, wi);
				si  = sat_add(si, tmp);
				sr = tre;
			}

			f0 += frame;
			f2 += frame;
		}
	}

	return 0;
}

